#import "NC2010AppDelegate.h"

@interface SmarterURLConnection : NSURLConnection {
	NSString* m_urlString ;
}

@property (copy) NSString* urlString ;

@end

@implementation SmarterURLConnection

@synthesize urlString = m_urlString ;

- (id)initWithRequest:(NSURLRequest*)request
			 delegate:(id)delegate {
	self = [super initWithRequest:request
						 delegate:delegate] ;
	if (self) {
		[self setUrlString:[[request URL] absoluteString]] ;
	}
	
	return self ;
}

- (void)dealloc {
	[m_urlString release] ;
	
	[super dealloc] ;
}

@end


@implementation NC2010AppDelegate

@synthesize urlStrings = m_urlStrings;

- (void)dealloc {
	[m_urlStrings release] ;
	
	[super dealloc] ;
}

- (void)fillPipeline {
	if (m_nextUrlIndex < [[self urlStrings] count]) {
		while (m_nInPipeline < [maxConnectionsField intValue]) {
			if ([logLevelControl selectedTag] > 0) { 
				NSLog(@"Requesting %d/%d with %d active connections", m_nextUrlIndex, [[self urlStrings] count], m_nInPipeline) ;
			}
			NSString* urlString = [[self urlStrings] objectAtIndex:m_nextUrlIndex++] ;
			NSURL *url = [NSURL URLWithString:urlString];
			NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:url
																   cachePolicy:NSURLRequestReloadIgnoringLocalCacheData
															   timeoutInterval:30.0];
			[request setHTTPShouldHandleCookies:NO]; // This makes no difference
			NSURLConnection *connection = [[SmarterURLConnection alloc] initWithRequest:request
																			   delegate:self] ;
			if ([logLevelControl selectedTag] > 1) {
				NSLog(@"Created request %p to %@", request, urlString);
			}
			if (connection) {
				m_nInPipeline++ ;
			}
			else if ([logLevelControl selectedTag] > 0) { 
				NSLog(@"connection error: %@", url);
			}
		}
	}
	else if (m_nInPipeline == 0) {
		NSLog(@"***** All connections have completed! *****") ;
	}
	else {
		if ([logLevelControl selectedTag] > 0) { 
			NSLog(@"%d connections still open", m_nInPipeline) ;
		}
	}
}

- (void)discountConnection:(NSURLConnection*)connection {
	[connection cancel] ;
	[connection release] ;
	
	m_nInPipeline-- ;
	[self fillPipeline] ;
}

- (void)connection:(NSURLConnection *)connection
didReceiveResponse:(NSURLResponse *)response {
	if ([logLevelControl selectedTag] > 1) {
		NSLog(@"Connection %p received response", connection) ;
	}	
	[self discountConnection:connection] ;
}

- (void)connection:(NSURLConnection *)connection
	didReceiveData:(NSData *)data {
	if ([logLevelControl selectedTag] > 1) {
		NSLog(@"Connection %p received bytes (%d)", connection, [data length]) ;
	}	
}

- (void)connection:(NSURLConnection *)connection
  didFailWithError:(NSError *)error {
	if ([logLevelControl selectedTag] > 0) { 
		NSLog(
			  @"Connection: %p\n        with URL: %@\ndidFailWithError: %@",
			  connection,
			  [(SmarterURLConnection*)connection urlString],
			  error) ;
	}
	[self discountConnection:connection] ;
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection {
	if ([logLevelControl selectedTag] > 0) { 
		NSLog( @"Connection did finish loading: %p", connection );
	}
	[self discountConnection:connection] ;
}

- (NSURLRequest *)connection:(NSURLConnection *)connection
			 willSendRequest:(NSURLRequest *)request
			redirectResponse:(NSURLResponse *)redirectResponse {
	if ([logLevelControl selectedTag] > 1) {
		NSLog( @"Connection %p did redirect to %@", connection, [request URL]) ;
	}
	return request;
}

- (void)threadOn {
	@synchronized(self) {
		m_nThreads++ ;
		if ([logLevelControl selectedTag] > 0) { 
			NSLog(@"%d active URL connection threads (+)", m_nThreads) ;
		}
	}
}

- (void)threadOff {
	@synchronized(self) {
		m_nThreads-- ;
		if ([logLevelControl selectedTag] > 0) { 
			NSLog(@"%d active URL connection threads (-)", m_nThreads) ;
		}
		
		if (m_nThreads == 0) {
			NSLog(@"*** All NSURLConnection Threads have completed. ***") ;
		}
	}
}

- (void)loadUrlString:(NSString *)urlString {
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	
	[self threadOn] ;
	NSURL *url = [NSURL URLWithString:urlString];
	NSError *error;
	NSString *download = [NSString stringWithContentsOfURL:url usedEncoding:NULL error:&error];
	if (download) {
		if ([logLevelControl selectedTag] > 1) {
			NSLog( @"Did finish loading %@", urlString) ;
		}
	}
	else if ([logLevelControl selectedTag] > 0) { 
		NSLog(@"Failed loading %@\ndue to error: %@", urlString, error) ;
	}
	
	[self threadOff] ;
	[pool release] ;
}

- (IBAction)startTest:(id)sender {
	m_nextUrlIndex = 0 ;
	
	// Read urlStrings from file into ivar
	NSString* path = [[NSBundle mainBundle] pathForResource:@"urls"
													 ofType:@"txt"] ;
	NSError* error ;
	NSString* urlsString = [NSString stringWithContentsOfFile:path
													 encoding:NSASCIIStringEncoding
														error:&error] ;
	if (!urlsString) {
		NSLog(@"Could not read urls.txt file.  %@", error) ;
		exit(1) ;
	}	
	NSArray* urlStrings = [urlsString componentsSeparatedByString:@"\n"] ;
	NSLog(@"Parsed %d url strings from resource urls.txt", [urlStrings count]) ;
	[self setUrlStrings:urlStrings] ;
	
	// Begin test
	switch ([regimenMatrix selectedTag]) {
		case 0:
			// Asynchronous NSURLConnection Methods
			NSLog( @"Making connections using Asynchronous NSURLConnection Methods...") ;
			[self fillPipeline] ;
			break;
		case 1:
			// Threaded -[NSString stringWithContentsOfURL]
			NSLog(@"*** Spawning threads to run -[NSString stringWithContentsOfURL]...") ;
			NSInteger nThreads = 0 ;
			// System limit is probably several hundred processes, so ...
			NSInteger maxThreads = [maxConnectionsField intValue] ;
			for (NSString* urlString in [self urlStrings]) {
				if (nThreads > maxThreads) {
					break ;
				}
				
				[NSThread detachNewThreadSelector:@selector(loadUrlString:)
										 toTarget:self
									   withObject:urlString] ;
				nThreads++ ;
				if ([logLevelControl selectedTag] > 1) {
					NSLog(@"%d/%d threads have been created", nThreads, maxThreads) ;
				}	
				
			}
			NSLog(@"*** Done spawning.  Spawned %d threads", nThreads) ;
			break;
		default:
			break;
	}
}

- (IBAction)stopTest:(id)sender {
	[self setUrlStrings:nil] ;
	if (([regimenMatrix selectedTag]) != 0) {
		NSBeep() ;
		NSLog(@"\n\n\n***** Cannot stop threads currently in progress! *****") ;
	}
}

@end
